@@ -57,8 +57,7 @@ G_DEFINE_TYPE (TpStreamEngineStream, tp_stream_engine_stream, G_TYPE_OBJECT);
 #define DEBUG(stream, format, ...) \
   g_debug ("stream %d (%s) %s: " format, \
     stream->stream_id, \
-    STREAM_PRIVATE (stream)->media_type == FARSIGHT_MEDIA_TYPE_AUDIO ? \
-      "audio" : "video", \
+    (stream)->media_type == FARSIGHT_MEDIA_TYPE_AUDIO ? "audio" : "video", \
     G_STRFUNC, \
     ##__VA_ARGS__)
 
@@ -76,7 +75,6 @@ struct _TpStreamEngineStreamPrivate
   TpPropsIface *conn_props;
   TpConn *connection_proxy;
 
-  guint media_type;
   FarsightStream *fs_stream;
   guint state_changed_handler_id;
 
@@ -94,7 +92,6 @@ struct _TpStreamEngineStreamPrivate
   gboolean candidate_preparation_required;
 
 #ifdef MAEMO_OSSO_SUPPORT
-  gboolean media_server_disabled;
   DBusGProxy *media_server_proxy;
 #endif
 };
@@ -136,11 +133,13 @@ static void
 set_remote_codecs (DBusGProxy *proxy, GPtrArray *codecs, gpointer user_data);
 static void
 set_stream_playing (DBusGProxy *proxy, gboolean play, gpointer user_data);
+static void
+set_stream_sending (DBusGProxy *proxy, gboolean play, gpointer user_data);
 
 static void
 stop_stream (TpStreamEngineStream *self);
 
-static gboolean
+gboolean
 g_object_has_property (GObject *object, const gchar *property)
 {
   GObjectClass *klass;
@@ -168,6 +167,11 @@ _remove_video_sink (TpStreamEngineStream *stream, GstElement *sink)
   gst_bin_remove (GST_BIN (pipeline), sink);
 }
 
+#ifdef MAEMO_OSSO_SUPPORT
+static void
+media_server_proxy_cleanup (TpStreamEngineStream *self);
+#endif
+
 static void
 tp_stream_engine_stream_dispose (GObject *object)
 {
@@ -176,12 +180,7 @@ tp_stream_engine_stream_dispose (GObject *object)
 
 #ifdef MAEMO_OSSO_SUPPORT
   if (priv->media_server_proxy)
-    {
-      DBusGProxy *proxy = priv->media_server_proxy;
-
-      priv->media_server_proxy = NULL;
-      g_object_unref (proxy);
-    }
+    media_server_proxy_cleanup (stream);
 #endif
 
   if (priv->channel_path)
@@ -225,6 +224,9 @@ tp_stream_engine_stream_dispose (GObject *object)
       dbus_g_proxy_disconnect_signal (priv->stream_handler_proxy,
           "SetStreamPlaying", G_CALLBACK (set_stream_playing), stream);
 
+      dbus_g_proxy_disconnect_signal (priv->stream_handler_proxy,
+          "SetStreamSending", G_CALLBACK (set_stream_sending), stream);
+
       g_object_unref (priv->stream_handler_proxy);
       priv->stream_handler_proxy = NULL;
     }
@@ -291,11 +293,6 @@ check_start_stream (TpStreamEngineStream *stream)
 {
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (stream);
 
-#ifdef MAEMO_OSSO_SUPPORT
-  if (!priv->media_server_disabled)
-    return;
-#endif
-
   DEBUG (stream, "stream_start_scheduled = %d; stream_started = %d",
     priv->stream_start_scheduled, priv->stream_started);
 
@@ -304,32 +301,40 @@ check_start_stream (TpStreamEngineStream *stream)
       if (farsight_stream_get_state (priv->fs_stream) == FARSIGHT_STREAM_STATE_CONNECTED)
         {
           farsight_stream_start (priv->fs_stream);
-
-          if (priv->media_type == FARSIGHT_MEDIA_TYPE_VIDEO)
-            {
-              TpStreamEngine *engine = tp_stream_engine_get ();
-              GstElement *pipeline = tp_stream_engine_get_pipeline (engine);
-              gst_element_set_state (pipeline, GST_STATE_PLAYING);
-            }
-
           priv->stream_started = TRUE;
         }
      }
 }
 
+typedef struct _method_call_ctx method_call_ctx;
+
+struct _method_call_ctx
+{
+  TpStreamEngineStream *stream;
+  const gchar *method;
+};
+
 /* dummy callback handler for async calling calls with no return values */
 static void
-dummy_callback (DBusGProxy *proxy, GError *error, gpointer user_data)
+async_method_callback (DBusGProxy *proxy, GError *error, gpointer user_data)
 {
+  method_call_ctx *ctx = (method_call_ctx *) user_data;
+
   if (error)
-    g_critical ("%s calling %s", error->message, (char*)user_data);
+    {
+      g_warning ("Error calling %s: %s", ctx->method, error->message);
+      g_signal_emit (ctx->stream, signals[STREAM_ERROR], 0);
+      g_error_free (error);
+    }
+
+  g_slice_free (method_call_ctx, ctx);
 }
 
 static void
-state_changed (FarsightStream *stream,
-               FarsightStreamState state,
-               FarsightStreamDirection dir,
-               gpointer user_data)
+cb_fs_state_changed (FarsightStream *stream,
+                     FarsightStreamState state,
+                     FarsightStreamDirection dir,
+                     gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
@@ -350,21 +355,26 @@ state_changed (FarsightStream *stream,
 
   if (priv->stream_handler_proxy)
     {
+      method_call_ctx *ctx = g_slice_new0 (method_call_ctx);
+
+      ctx->stream = self;
+      ctx->method = "Media.StreamHandler::StreamState";
+
       tp_media_stream_handler_stream_state_async (
-        priv->stream_handler_proxy, state, dummy_callback,
-        "Media.StreamHandler::StreamState");
+        priv->stream_handler_proxy, state, async_method_callback, ctx);
     }
 }
 
 static void
-new_native_candidate (FarsightStream *stream,
-                      gchar *candidate_id,
-                      gpointer user_data)
+cb_fs_new_native_candidate (FarsightStream *stream,
+                            gchar *candidate_id,
+                            gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
   const GList *fs_candidates, *lp;
   GPtrArray *transports;
+  method_call_ctx *ctx;
 
   fs_candidates = farsight_stream_get_native_candidate (stream, candidate_id);
   transports = g_ptr_array_new ();
@@ -427,9 +437,13 @@ new_native_candidate (FarsightStream *stream,
       g_ptr_array_add (transports, g_value_get_boxed (&transport));
     }
 
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NativeCandidatesPrepared";
+
   tp_media_stream_handler_new_native_candidate_async (
-      priv->stream_handler_proxy, candidate_id, transports, dummy_callback,
-      "Media.StreamHandler::NativeCandidatesPrepared");
+      priv->stream_handler_proxy, candidate_id, transports,
+      async_method_callback, ctx);
 }
 
 /**
@@ -446,7 +460,7 @@ tp_transports_to_fs (gchar* candidate, GPtrArray *transports)
   GList *fs_trans_list = NULL;
   GValueArray *transport;
   FarsightTransportInfo *fs_transport;
-  int i;
+  guint i;
 
   for (i=0; i< transports->len; i++)
     {
@@ -620,7 +634,7 @@ set_remote_candidate_list (DBusGProxy *proxy, GPtrArray *candidates,
   GValueArray *candidate = NULL;
   GPtrArray *transports = NULL;
   gchar *candidate_id = NULL;
-  int i;
+  guint i;
 
   for (i = 0; i < candidates->len; i++)
     {
@@ -663,8 +677,9 @@ set_remote_codecs (DBusGProxy *proxy, GPtrArray *codecs, gpointer user_data)
   GHashTable *params = NULL;
   FarsightCodec *fs_codec;
   GList *fs_params = NULL;
-  int i;
+  guint i;
   GPtrArray *supp_codecs;
+  method_call_ctx *ctx;
 
   DEBUG (self, "called");
 
@@ -711,9 +726,12 @@ set_remote_codecs (DBusGProxy *proxy, GPtrArray *codecs, gpointer user_data)
   supp_codecs = fs_codecs_to_tp (
       farsight_stream_get_codec_intersection (priv->fs_stream));
 
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::SupportedCodecs";
+
   tp_media_stream_handler_supported_codecs_async
-    (priv->stream_handler_proxy, supp_codecs, dummy_callback,
-     "Media.StreamHandler::SupportedCodecs");
+    (priv->stream_handler_proxy, supp_codecs, async_method_callback, ctx);
 
   for (lp = g_list_first (fs_codecs); lp; lp = g_list_next (lp))
     {
@@ -741,11 +759,11 @@ stop_stream (TpStreamEngineStream *self)
 
   DEBUG (self, "calling stop on farsight stream %p", priv->fs_stream);
 
-  if (priv->media_type == FARSIGHT_MEDIA_TYPE_VIDEO)
+  if (self->media_type == FARSIGHT_MEDIA_TYPE_VIDEO)
     {
       GstElement *sink = farsight_stream_get_sink (priv->fs_stream);
-      farsight_stream_stop (priv->fs_stream);
       _remove_video_sink (self, sink);
+      farsight_stream_stop (priv->fs_stream);
     }
   else
     {
@@ -755,7 +773,7 @@ stop_stream (TpStreamEngineStream *self)
   priv->stream_started = FALSE;
 
 #ifdef MAEMO_OSSO_SUPPORT
-  if (priv->media_server_disabled && priv->media_server_proxy)
+  if (priv->media_server_proxy != NULL)
     {
       GError *error = NULL;
 
@@ -768,9 +786,9 @@ stop_stream (TpStreamEngineStream *self)
         g_message ("unable to re-enable media server: %s", error->message);
         g_error_free (error);
       }
-    }
 
-  priv->media_server_disabled = FALSE;
+      media_server_proxy_cleanup (self);
+    }
 #endif
 
 }
@@ -796,11 +814,26 @@ set_stream_playing (DBusGProxy *proxy, gboolean play, gpointer user_data)
     }
 }
 
+static void
+set_stream_sending (DBusGProxy *proxy, gboolean send, gpointer user_data)
+{
+  TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
+  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+
+  g_assert (priv->fs_stream != NULL);
+
+  DEBUG (self, "%d", send);
+
+  farsight_stream_set_sending (priv->fs_stream, send);
+}
+
 static void
 close (DBusGProxy *proxy, gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
 
+  DEBUG (self, "close requested by connection manager");
+
   stop_stream (self);
   g_signal_emit (self, signals[STREAM_CLOSED], 0);
 }
@@ -814,23 +847,33 @@ prepare_transports (TpStreamEngineStream *self)
   if (priv->got_connection_properties && priv->candidate_preparation_required)
     {
       farsight_stream_prepare_transports (priv->fs_stream);
+      method_call_ctx *ctx;
+
 
       codecs = fs_codecs_to_tp (
                  farsight_stream_get_local_codecs (priv->fs_stream));
 
       DEBUG (self, "calling MediaStreamHandler::Ready");
+
+      ctx = g_slice_new0 (method_call_ctx);
+      ctx->stream = self;
+      ctx->method = "Media.StreamHandler::Ready";
+
       tp_media_stream_handler_ready_async (
-        priv->stream_handler_proxy, codecs, dummy_callback, self);
+        priv->stream_handler_proxy, codecs, async_method_callback, ctx);
     }
 }
 
 static void
-codec_changed (FarsightStream *stream, gint codec_id, gpointer user_data)
+cb_fs_codec_changed (FarsightStream *stream,
+                     gint codec_id,
+                     gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+  method_call_ctx *ctx;
 
-  if (priv->media_type == FARSIGHT_MEDIA_TYPE_AUDIO)
+  if (self->media_type == FARSIGHT_MEDIA_TYPE_AUDIO)
     {
       tp_stream_engine_stream_mute_output (self, priv->output_mute, NULL);
       tp_stream_engine_stream_mute_input (self, priv->input_mute, NULL);
@@ -842,46 +885,72 @@ codec_changed (FarsightStream *stream, gint codec_id, gpointer user_data)
       self->stream_id, TRUE);
 
   DEBUG (self, "codec_id=%d, stream=%p", codec_id, stream);
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::CodecChoice";
+
   tp_media_stream_handler_codec_choice_async (
-    priv->stream_handler_proxy, codec_id, dummy_callback,
-    "Media.StreamHandler::CodecChoice");
+    priv->stream_handler_proxy, codec_id, async_method_callback, ctx);
 }
 
-static void
-stream_error (
-  FarsightStream *stream,
-  FarsightStreamError error,
-  const gchar *debug,
-  gpointer user_data)
+void
+tp_stream_engine_stream_error (
+  TpStreamEngineStream *self,
+  guint error,
+  const gchar *message)
 {
-  TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
 
-  g_message ("%s: stream error: stream=%p error=%s", G_STRFUNC, stream, debug);
-  /* FIXME: check if error is EOS */
-  tp_media_stream_handler_error (priv->stream_handler_proxy, 0, debug, NULL);
+  g_message ("%s: stream errorno=%d error=%s", G_STRFUNC, error, message);
+
+  tp_media_stream_handler_error (priv->stream_handler_proxy, error,
+      message, NULL);
   g_signal_emit (self, signals[STREAM_ERROR], 0);
 }
 
 static void
-new_active_candidate_pair (FarsightStream *stream, const gchar* native_candidate, const gchar *remote_candidate, gpointer user_data)
+cb_fs_stream_error (FarsightStream *stream,
+                    FarsightStreamError error,
+                    const gchar *debug,
+                    gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
-  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
 
-  DEBUG (self, "stream=%p", stream);
-
-  tp_media_stream_handler_new_active_candidate_pair_async
-    (priv->stream_handler_proxy, native_candidate, remote_candidate, dummy_callback, "Media.StreamHandler::NewActiveCandidatePair");
+  /* FIXME: map Farsight errors to Telepathy errors */
+  tp_stream_engine_stream_error (self, 0, debug);
 }
 
 static void
-native_candidates_prepared (FarsightStream *stream, gpointer user_data)
+cb_fs_new_active_candidate_pair (FarsightStream *stream,
+                                 const gchar* native_candidate,
+                                 const gchar *remote_candidate,
+                                 gpointer user_data)
+{
+  TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
+  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+  method_call_ctx *ctx;
+
+  DEBUG (self, "stream=%p", stream);
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NewActiveCandidatePair";
+
+  tp_media_stream_handler_new_active_candidate_pair_async (
+    priv->stream_handler_proxy, native_candidate, remote_candidate,
+    async_method_callback, ctx);
+}
+
+static void
+cb_fs_native_candidates_prepared (FarsightStream *stream,
+                                  gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
   const GList *transport_candidates, *lp;
   FarsightTransportInfo *info;
+  method_call_ctx *ctx;
 
   DEBUG (self, "stream=%p", stream);
 
@@ -894,9 +963,13 @@ native_candidates_prepared (FarsightStream *stream, gpointer user_data)
         (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "TCP" : "UDP",
         info->proto_subtype, info->ip, info->port, (double) info->preference);
   }
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NativeCandidatesPrepared";
+
   tp_media_stream_handler_native_candidates_prepared_async (
-    priv->stream_handler_proxy, dummy_callback,
-    "Media.StreamHandler::NativeCandidatesPrepared");
+    priv->stream_handler_proxy, async_method_callback, ctx);
 }
 
 static void
@@ -945,24 +1018,30 @@ cb_properties_ready (TpPropsIface *iface, gpointer user_data)
 }
 
 #ifdef MAEMO_OSSO_SUPPORT
+static void
+media_server_proxy_cleanup (TpStreamEngineStream *self)
+{
+  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+  DBusGProxy *proxy;
+
+  if (priv->media_server_proxy == NULL)
+    return;
+
+  proxy = priv->media_server_proxy;
+  priv->media_server_proxy = NULL;
+  g_object_unref (proxy);
+}
+
 static void
 media_server_proxy_destroyed (DBusGProxy *proxy, gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
-  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
 
-  if (priv->media_server_proxy)
-    {
-      DBusGProxy *proxy = priv->media_server_proxy;
-
-      DEBUG (self, "media server proxy destroyed");
-
-      priv->media_server_proxy = NULL;
-      g_object_unref (proxy);
-    }
+  DEBUG (self, "media server proxy destroyed");
+  media_server_proxy_cleanup (self);
 }
 
-static gboolean
+static void
 media_server_proxy_init (TpStreamEngineStream *self)
 {
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
@@ -981,22 +1060,21 @@ media_server_proxy_init (TpStreamEngineStream *self)
 
   g_message ("disabling media server");
 
-  if (com_nokia_osso_media_server_disable (
+  if (!com_nokia_osso_media_server_disable (
         DBUS_G_PROXY (priv->media_server_proxy),
         &me_error))
-    {
-      priv->media_server_disabled = TRUE;
-      return TRUE;
-    }
-  else
     {
       if (me_error)
-        g_message ("failed to disable media server: %s", me_error->message);
+        {
+          g_message ("failed to disable media server: %s", me_error->message);
+          g_error_free (me_error);
+        }
       else
-        g_message ("failed to disable media server");
+        {
+          g_message ("failed to disable media server");
+        }
 
-      priv->media_server_disabled = FALSE;
-      return FALSE;
+      media_server_proxy_cleanup (self);
     }
 }
 #endif
@@ -1071,6 +1149,7 @@ make_sink (TpStreamEngineStream *stream, guint media_type)
           TpStreamEngine *engine;
           DEBUG (stream, "making video sink with pipeline \"%s\"", elem);
           sink = gst_parse_bin_from_description (elem, TRUE, NULL);
+          g_assert (GST_IS_BIN (sink));
           engine = tp_stream_engine_get ();
           gst_bin_add (GST_BIN (tp_stream_engine_get_pipeline (engine)), sink);
           g_assert (sink);
@@ -1109,14 +1188,14 @@ tp_stream_engine_stream_go (
   gchar *conn_timeout_str;
 
 #ifdef MAEMO_OSSO_SUPPORT
-  if (!media_server_proxy_init (stream))
-    return FALSE;
+  if (media_type == TP_MEDIA_STREAM_TYPE_AUDIO)
+    media_server_proxy_init (stream);
 #endif
 
   priv->channel_path = g_strdup (channel_path);
 
   stream->stream_id = id;
-  priv->media_type = media_type;
+  stream->media_type = media_type;
   priv->stream_handler_proxy = dbus_g_proxy_new_for_name (
     tp_get_bus(),
     bus_name,
@@ -1176,18 +1255,18 @@ tp_stream_engine_stream_go (
     }
 
   g_signal_connect (G_OBJECT (priv->fs_stream), "error",
-                    G_CALLBACK (stream_error), stream);
+                    G_CALLBACK (cb_fs_stream_error), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "new-active-candidate-pair",
-                    G_CALLBACK (new_active_candidate_pair), stream);
+                    G_CALLBACK (cb_fs_new_active_candidate_pair), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "codec-changed",
-                    G_CALLBACK (codec_changed), stream);
+                    G_CALLBACK (cb_fs_codec_changed), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "native-candidates-prepared",
-                    G_CALLBACK (native_candidates_prepared), stream);
+                    G_CALLBACK (cb_fs_native_candidates_prepared), stream);
   priv->state_changed_handler_id =
     g_signal_connect (G_OBJECT (priv->fs_stream), "state-changed",
-                      G_CALLBACK (state_changed), stream);
+                      G_CALLBACK (cb_fs_state_changed), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "new-native-candidate",
-                    G_CALLBACK (new_native_candidate), stream);
+                    G_CALLBACK (cb_fs_new_native_candidate), stream);
 
   /* OMG, Can we make dbus-binding-tool do this stuff for us?? */
 
@@ -1203,6 +1282,8 @@ tp_stream_engine_stream_go (
       TP_TYPE_CODEC_LIST, G_TYPE_INVALID);
   dbus_g_proxy_add_signal (priv->stream_handler_proxy, "SetStreamPlaying",
       G_TYPE_BOOLEAN, G_TYPE_INVALID);
+  dbus_g_proxy_add_signal (priv->stream_handler_proxy, "SetStreamSending",
+      G_TYPE_BOOLEAN, G_TYPE_INVALID);
   dbus_g_proxy_add_signal (priv->stream_handler_proxy, "Close",
       G_TYPE_INVALID);
 
@@ -1218,14 +1299,13 @@ tp_stream_engine_stream_go (
       G_CALLBACK (set_remote_codecs), stream, NULL);
   dbus_g_proxy_connect_signal (priv->stream_handler_proxy, "SetStreamPlaying",
       G_CALLBACK (set_stream_playing), stream, NULL);
+  dbus_g_proxy_connect_signal (priv->stream_handler_proxy, "SetStreamSending",
+      G_CALLBACK (set_stream_sending), stream, NULL);
   dbus_g_proxy_connect_signal (priv->stream_handler_proxy, "Close",
       G_CALLBACK (close), stream, NULL);
 
   priv->candidate_preparation_required = TRUE;
 
-  set_stun (stream);
-  prepare_transports (stream);
-
   priv->connection_proxy = tp_conn_new (
     tp_get_bus(),
     bus_name,
@@ -1245,22 +1325,31 @@ tp_stream_engine_stream_go (
    */
   priv->conn_props = TELEPATHY_PROPS_IFACE (tp_conn_get_interface (
         priv->connection_proxy, TELEPATHY_PROPS_IFACE_QUARK));
-  g_assert (priv->conn_props);
 
-  /* surely we don't need all of these properties */
-  tp_props_iface_set_mapping (priv->conn_props,
-      "stun-server", CONN_PROP_STUN_SERVER,
-      "stun-port", CONN_PROP_STUN_PORT,
-      "stun-relay-server", CONN_PROP_STUN_RELAY_SERVER,
-      "stun-relay-udp-port", CONN_PROP_STUN_RELAY_UDP_PORT,
-      "stun-relay-tcp-port", CONN_PROP_STUN_RELAY_TCP_PORT,
-      "stun-relay-ssltcp-port", CONN_PROP_STUN_RELAY_SSLTCP_PORT,
-      "stun-relay-username", CONN_PROP_STUN_RELAY_USERNAME,
-      "stun-relay-password", CONN_PROP_STUN_RELAY_PASSWORD,
-      NULL);
+  if (priv->conn_props != NULL)
+    {
+      /* surely we don't need all of these properties */
+      tp_props_iface_set_mapping (priv->conn_props,
+          "stun-server", CONN_PROP_STUN_SERVER,
+          "stun-port", CONN_PROP_STUN_PORT,
+          "stun-relay-server", CONN_PROP_STUN_RELAY_SERVER,
+          "stun-relay-udp-port", CONN_PROP_STUN_RELAY_UDP_PORT,
+          "stun-relay-tcp-port", CONN_PROP_STUN_RELAY_TCP_PORT,
+          "stun-relay-ssltcp-port", CONN_PROP_STUN_RELAY_SSLTCP_PORT,
+          "stun-relay-username", CONN_PROP_STUN_RELAY_USERNAME,
+          "stun-relay-password", CONN_PROP_STUN_RELAY_PASSWORD,
+          NULL);
 
-  g_signal_connect (priv->conn_props, "properties-ready",
-                    G_CALLBACK (cb_properties_ready), stream);
+      g_signal_connect (priv->conn_props, "properties-ready",
+                        G_CALLBACK (cb_properties_ready), stream);
+    }
+  else
+    {
+      priv->got_connection_properties = TRUE;
+    }
+
+  set_stun (stream);
+  prepare_transports (stream);
 
   return TRUE;
 }
@@ -1281,7 +1370,7 @@ gboolean tp_stream_engine_stream_mute_output (
 
   g_return_val_if_fail (priv->fs_stream, FALSE);
 
-  if (priv->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
+  if (stream->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
     {
       *error = g_error_new (TELEPATHY_ERRORS, InvalidArgument,
         "MuteInput can only be called on audio streams");
@@ -1311,7 +1400,7 @@ gboolean tp_stream_engine_stream_set_output_volume (
 
   g_return_val_if_fail (priv->fs_stream, FALSE);
 
-  if (priv->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
+  if (stream->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
     {
       *error = g_error_new (TELEPATHY_ERRORS, InvalidArgument,
         "SetOutputVolume can only be called on audio streams");
@@ -1342,7 +1431,7 @@ gboolean tp_stream_engine_stream_mute_input (
 
   g_return_val_if_fail (priv->fs_stream, FALSE);
 
-  if (priv->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
+  if (stream->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
     {
       *error = g_error_new (TELEPATHY_ERRORS, InvalidArgument,
         "MuteInput can only be called on audio streams");
@@ -1369,10 +1458,9 @@ tp_stream_engine_stream_set_output_window (
 {
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (stream);
   TpStreamEngine *engine;
-  GstElement *pipeline, *sink;
-  const gchar *videosink_name;
+  GstElement *sink;
 
-  if (priv->media_type != FARSIGHT_MEDIA_TYPE_VIDEO)
+  if (stream->media_type != FARSIGHT_MEDIA_TYPE_VIDEO)
     {
       DEBUG (stream, "can only be called on video streams");
       *error = g_error_new (TELEPATHY_ERRORS, InvalidArgument,
@@ -1407,24 +1495,20 @@ tp_stream_engine_stream_set_output_window (
       return TRUE;
     }
 
+  sink = tp_stream_engine_make_video_sink (engine);
+
+  if (sink == NULL)
+    {
+      DEBUG (stream, "failed to make video sink, no output for window %d :(",
+          window_id);
+      g_set_error (error, TELEPATHY_ERRORS, NotAvailable, "failed to make a "
+          "video sink");
+      return FALSE;
+    }
+
   DEBUG (stream, "putting video output in window %d", window_id);
 
-  if ((videosink_name = getenv ("FS_VIDEO_SINK")) || (videosink_name = getenv("FS_VIDEOSINK")))
-    {
-      DEBUG (stream, "making video sink with pipeline \"%s\"", videosink_name);
-      sink = gst_parse_bin_from_description (videosink_name, TRUE, NULL);
-    }
-  else
-    {
-      DEBUG (stream, "using xvimagesink");
-      sink = gst_element_factory_make ("xvimagesink", NULL);
-      g_object_set (sink, "sync", FALSE, NULL);
-    }
-
-  pipeline = tp_stream_engine_get_pipeline (engine);
-  gst_bin_add (GST_BIN (pipeline), sink);
   tp_stream_engine_add_output_window (engine, stream, sink, window_id);
-
   farsight_stream_set_sink (priv->fs_stream, sink);
   gst_object_unref (sink);
 
