@@ -107,6 +107,7 @@ enum
   HANDLING_CHANNEL,
   NO_MORE_CHANNELS,
   RECEIVING,
+  SHUTDOWN_REQUESTED,
   LAST_SIGNAL
 };
 
@@ -119,6 +120,7 @@ struct _TpStreamEnginePrivate
   gboolean dispose_has_run;
 
   GstElement *pipeline;
+  gboolean pipeline_playing;
 
   GPtrArray *channels;
 
@@ -147,6 +149,7 @@ struct _WindowPair
   GstElement *sink;
   guint window_id;
   volatile gboolean removing;
+  gboolean created;
 };
 
 static void
@@ -166,6 +169,7 @@ _window_pairs_free (GSList **list)
 static void
 _window_pairs_add (GSList **list, TpStreamEngineStream *stream, GstElement *sink, guint window_id)
 {
+  g_debug ("Adding to windowpair list sink %p window_id %d", sink, window_id);
   WindowPair *wp;
 
   wp = g_slice_new (WindowPair);
@@ -173,6 +177,7 @@ _window_pairs_add (GSList **list, TpStreamEngineStream *stream, GstElement *sink
   wp->sink = sink;
   wp->window_id = window_id;
   wp->removing = FALSE;
+  wp->created = FALSE;
 
   *list = g_slist_prepend (*list, wp);
 }
@@ -204,6 +209,29 @@ _window_pairs_find_by_removing (GSList *list, gboolean removing)
   return NULL;
 }
 
+static gboolean
+_gst_bin_find_element (GstBin *bin, GstElement *element)
+{
+  GstIterator *iter;
+  gpointer item;
+  gboolean found = FALSE;
+
+  iter = gst_bin_iterate_elements (bin);
+
+  while (!found && gst_iterator_next (iter, &item) == GST_ITERATOR_OK)
+    {
+      GstElement *child = (GstElement *) item;
+
+      if (child == element)
+        found = TRUE;
+
+      gst_object_unref (child);
+    }
+
+  gst_iterator_free (iter);
+  return found;
+}
+
 static WindowPair *
 _window_pairs_find_by_sink (GSList *list, GstElement *sink)
 {
@@ -216,6 +244,9 @@ _window_pairs_find_by_sink (GSList *list, GstElement *sink)
       WindowPair *wp = tmp->data;
       if (wp->sink == sink)
         return wp;
+      else if (GST_IS_BIN (wp->sink))
+        if (_gst_bin_find_element (GST_BIN (wp->sink), GST_ELEMENT (sink)))
+          return wp;
     }
 
   return NULL;
@@ -238,6 +269,188 @@ _window_pairs_find_by_window_id (GSList *list, guint window_id)
   return NULL;
 }
 
+GstElement *
+tp_stream_engine_make_video_sink (TpStreamEngine *obj)
+{
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
+  const gchar *videosink_name;
+  GstElement *sink = NULL;
+#ifndef MAEMO_OSSO_SUPPORT
+  GstElement *bin, *tmp;
+  GstPad *pad;
+#endif
+
+  g_assert (priv->pipeline != NULL);
+
+  if ((videosink_name = getenv ("FS_VIDEO_SINK")) || (videosink_name = getenv("FS_VIDEOSINK")))
+    {
+      g_debug ("making video sink with pipeline \"%s\"", videosink_name);
+      sink = gst_parse_bin_from_description (videosink_name, TRUE, NULL);
+    }
+  else
+    {
+#ifndef MAEMO_OSSO_SUPPORT
+      sink = gst_element_factory_make ("gconfvideosink", NULL);
+
+      if (sink == NULL)
+        sink = gst_element_factory_make ("autovideosink", NULL);
+#endif
+
+      if (sink == NULL)
+        sink = gst_element_factory_make ("xvimagesink", NULL);
+
+#ifndef MAEMO_OSSO_SUPPORT
+      if (sink == NULL)
+        sink = gst_element_factory_make ("ximagesink", NULL);
+#endif
+    }
+
+  if (sink != NULL)
+    {
+      g_debug ("made video sink element %s", GST_ELEMENT_NAME (sink));
+
+      if (g_object_has_property (G_OBJECT (sink), "sync"))
+        {
+          g_debug ("setting sync to FALSE");
+          g_object_set (G_OBJECT (sink), "sync", FALSE, NULL);
+        }
+    }
+  else
+    {
+      g_debug ("failed to make a video sink");
+      return NULL;
+    }
+
+#ifndef MAEMO_OSSO_SUPPORT
+  bin = gst_bin_new (NULL);
+  gst_bin_add (GST_BIN (bin), sink);
+
+  tmp = gst_element_factory_make ("videoscale", NULL);
+  if (tmp != NULL)
+    {
+      g_debug ("linking videoscale");
+      gst_bin_add (GST_BIN (bin), tmp);
+      gst_element_link (tmp, sink);
+      sink = tmp;
+    }
+
+  tmp = gst_element_factory_make ("ffmpegcolorspace", NULL);
+  if (tmp != NULL);
+    {
+      g_debug ("linking ffmpegcolorspace");
+      gst_bin_add (GST_BIN (bin), tmp);
+      gst_element_link (tmp, sink);
+      sink = tmp;
+    }
+
+  pad = gst_bin_find_unconnected_pad (GST_BIN (bin), GST_PAD_SINK);
+  gst_element_add_pad (bin, gst_ghost_pad_new ("sink", pad));
+  gst_object_unref (GST_OBJECT (pad));
+
+  sink = bin;
+#endif
+
+  gst_bin_add (GST_BIN (priv->pipeline), sink);
+
+  return sink;
+}
+
+static gboolean
+_add_preview_window (TpStreamEngine *obj, guint window_id, GError **error)
+{
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
+  WindowPair *wp;
+  GstElement *tee, *sink;
+  GstStateChangeReturn state_change_ret;
+
+  g_assert (priv->pipeline != NULL);
+
+  g_debug ("%s: called for window id %d", G_STRFUNC, window_id);
+  wp = _window_pairs_find_by_window_id (priv->preview_windows, window_id);
+
+  if (wp == NULL)
+    {
+      *error = g_error_new (TELEPATHY_ERRORS, InvalidArgument,
+          "Given window id %d not in windowID list", window_id);
+      g_debug ("%s: Couldn't find xwindow id in window pair list", G_STRFUNC);
+      return FALSE;
+    }
+
+  g_debug ("adding preview in window %u", window_id);
+
+  tee = gst_bin_get_by_name (GST_BIN (priv->pipeline), "tee");
+  sink = tp_stream_engine_make_video_sink (obj);
+
+  if (sink == NULL)
+    goto sink_failure;
+
+  wp->created = TRUE;
+  wp->sink = sink;
+
+  g_debug ("trying to set sink to PLAYING");
+  state_change_ret = gst_element_set_state (sink, GST_STATE_PLAYING);
+
+  if (state_change_ret != GST_STATE_CHANGE_SUCCESS &&
+      state_change_ret != GST_STATE_CHANGE_NO_PREROLL &&
+      state_change_ret != GST_STATE_CHANGE_ASYNC)
+    {
+      g_set_error (error, GST_STREAM_ERROR, GST_STREAM_ERROR_FAILED,
+          "Failed to set element %s to PLAYING", GST_ELEMENT_NAME (sink));
+      goto link_failure;
+    }
+
+  g_debug ("trying to link tee and sink");
+  if (!gst_element_link (tee, sink))
+    {
+      g_set_error (error, GST_STREAM_ERROR, GST_STREAM_ERROR_FAILED,
+          "Failed to link element %s to %s", GST_ELEMENT_NAME (tee),
+          GST_ELEMENT_NAME (sink));
+      goto link_failure;
+    }
+
+  gst_object_unref (tee);
+  g_signal_emit (obj, signals[HANDLING_CHANNEL], 0);
+  return TRUE;
+
+link_failure:
+  gst_element_set_state (sink, GST_STATE_NULL);
+  gst_bin_remove (GST_BIN (priv->pipeline), sink);
+
+sink_failure:
+  gst_object_unref (tee);
+
+  if (error != NULL)
+    g_warning ((*error)->message);
+
+  _window_pairs_remove (&(priv->preview_windows), wp);
+
+  return FALSE;
+}
+
+static void
+_add_pending_preview_windows (TpStreamEngine *engine)
+{
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (engine);
+
+  g_debug ("%s: called", G_STRFUNC);
+
+  GSList *tmp;
+
+  for (tmp = priv->preview_windows;
+       tmp != NULL;
+       tmp = tmp->next)
+    {
+      WindowPair *wp = tmp->data;
+      if (wp->created == FALSE)
+        {
+          GError *error = NULL;
+          _add_preview_window (engine, wp->window_id, &error);
+
+          if (error != NULL)
+            g_error_free (error);
+        }
+    }
+}
 
 #ifdef USE_INFOPRINT
 static void
@@ -372,6 +585,20 @@ tp_stream_engine_class_init (TpStreamEngineClass *tp_stream_engine_class)
         tp_stream_engine_marshal_VOID__STRING_INT_BOOLEAN,
         G_TYPE_NONE, 3, DBUS_TYPE_G_OBJECT_PATH, G_TYPE_UINT, G_TYPE_BOOLEAN);
 
+  /**
+   * TpStreamEngine::shutdown:
+   *
+   * Emitted whenever stream engine needs to be shutdown
+   */
+  signals[SHUTDOWN_REQUESTED] =
+    g_signal_new ("shutdown-requested",
+        G_OBJECT_CLASS_TYPE (tp_stream_engine_class),
+        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
+        0,
+        NULL, NULL,
+        g_cclosure_marshal_VOID__VOID,
+        G_TYPE_NONE, 0);
+
   dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (tp_stream_engine_class), &dbus_glib_tp_stream_engine_object_info);
 }
 
@@ -522,7 +749,7 @@ channel_closed (TpStreamEngineChannel *chan, gpointer user_data)
 
 
 static void
-_remove_defunct_sinks (TpStreamEngine *engine)
+_remove_defunct_preview_sinks (TpStreamEngine *engine, gboolean clear_wp_list)
 {
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (engine);
   WindowPair *wp = NULL;
@@ -535,16 +762,50 @@ _remove_defunct_sinks (TpStreamEngine *engine)
       g_debug ("%s: removing sink for preview window ID %u", G_STRFUNC,
           wp->window_id);
 
-      tee = gst_bin_get_by_name (GST_BIN (priv->pipeline), "tee");
-      gst_element_unlink (tee, wp->sink);
-      gst_object_unref (GST_OBJECT (tee));
+      if (wp->created)
+        {
+          g_assert (wp->sink != NULL);
+          tee = gst_bin_get_by_name (GST_BIN (priv->pipeline), "tee");
+          g_assert (tee != NULL);
+          gst_element_unlink (tee, wp->sink);
+          gst_object_unref (GST_OBJECT (tee));
+          g_debug ("unlinked sink from tee, now setting state to NULL");
 
-      gst_element_set_state (wp->sink, GST_STATE_NULL);
-      gst_bin_remove (GST_BIN (priv->pipeline), wp->sink);
+          gst_element_set_state (wp->sink, GST_STATE_NULL);
+          g_debug ("Done setting state to NULL, now removing from bin");
+          gst_bin_remove (GST_BIN (priv->pipeline), wp->sink);
+          g_debug ("Done removing from bin, calling _window_pairs_remove"
+              "refcount %d", GST_OBJECT_REFCOUNT_VALUE (wp->sink));
+        }
+      else
+        {
+          g_debug ("No sink created yet, removing window_pair");
+          g_assert (wp->sink == NULL);
+        }
 
-      _window_pairs_remove (&(priv->preview_windows), wp);
+      if (clear_wp_list)
+        {
+          _window_pairs_remove (&(priv->preview_windows), wp);
+        }
+      else
+        {
+          wp->sink = NULL;
+          wp->created = FALSE;
+          wp->stream = NULL;
+          wp->removing = FALSE;
+        }
+      g_debug ("Done _window_pairs_remove");
     }
 
+    check_if_busy (engine);
+}
+
+static void
+_remove_defunct_output_sinks (TpStreamEngine *engine)
+{
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (engine);
+  WindowPair *wp = NULL;
+
   while ((wp = _window_pairs_find_by_removing (priv->output_windows, TRUE)) !=
       NULL)
     {
@@ -563,6 +824,38 @@ _remove_defunct_sinks (TpStreamEngine *engine)
   check_if_busy (engine);
 }
 
+
+static void
+close_all_video_streams (TpStreamEngine *self, const gchar *message)
+{
+  g_debug ("Closing all video streams");
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (self);
+  guint i, j, k;
+
+  for (i = 0; i < priv->channels->len; i++)
+    {
+      TpStreamEngineChannel *channel = g_ptr_array_index (
+            priv->channels, i);
+
+      for (j = 0; j < channel->sessions->len; j++)
+        {
+          TpStreamEngineSession *session = g_ptr_array_index (
+              channel->sessions, j);
+
+          for (k = 0; k < session->streams->len; k++)
+            {
+              TpStreamEngineStream *stream = g_ptr_array_index (
+                session->streams, k);
+
+              if (stream->media_type == TP_MEDIA_STREAM_TYPE_VIDEO)
+                tp_stream_engine_stream_error (stream,
+                    TP_MEDIA_STREAM_ERROR_UNKNOWN, message);
+            }
+        }
+    }
+}
+
+
 static gboolean
 bus_async_handler (GstBus *bus,
                    GstMessage *message,
@@ -572,37 +865,145 @@ bus_async_handler (GstBus *bus,
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (engine);
   GError *error = NULL;
   WindowPair *wp;
+  gchar *error_string;
 
-  if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ERROR)
-    return TRUE;
+  GstElement *source = GST_ELEMENT (GST_MESSAGE_SRC (message));
+  gchar *name = gst_element_get_name (source);
 
-  gst_message_parse_error (message, &error, NULL);
-
-  g_debug ("%s: got error: %s", G_STRFUNC, error->message);
-
-  if (error->domain == GST_RESOURCE_ERROR)
+  switch (GST_MESSAGE_TYPE (message))
     {
-      if (error->code == GST_RESOURCE_ERROR_WRITE)
-        {
-          wp = _window_pairs_find_by_sink (priv->preview_windows,
-              GST_ELEMENT (GST_MESSAGE_SRC (message)));
+      case GST_MESSAGE_ERROR:
+        gst_message_parse_error (message, &error, &error_string);
 
-          if (wp == NULL)
-            wp = _window_pairs_find_by_sink (priv->output_windows,
-                GST_ELEMENT (GST_MESSAGE_SRC (message)));
+        g_debug ("%s: got error from %s", G_STRFUNC, name);
+        g_debug ("%s: got error: %s %s %d %d", G_STRFUNC, error->message,
+            error_string, error->domain, error->code);
+        g_free (error_string);
 
-          if (wp != NULL)
+        /*
+        TpStreamEngineStream *stream = NULL;
+        guint xid;
+        gboolean is_preview = TRUE;
+        */
+        if (g_strrstr (name, "xvimagesink") &&
+            error->domain == GST_RESOURCE_ERROR &&
+            (error->code == GST_RESOURCE_ERROR_WRITE ||
+             error->code == GST_RESOURCE_ERROR_BUSY))
+          {
+            g_debug ("%s: error from xvimagesink, shutting down video streams",
+                G_STRFUNC);
+            wp = _window_pairs_find_by_sink (priv->preview_windows,
+                source);
+
+            if (wp == NULL)
             {
-              g_debug ("%s: sink for %s window (id %u) has gone, removing",
-                  G_STRFUNC, wp->stream == NULL ? "preview" : "output",
-                  wp->window_id);
-
-              wp->removing = TRUE;
-              _remove_defunct_sinks (engine);
+              /* is_preview = FALSE; */
+              wp = _window_pairs_find_by_sink (priv->output_windows,
+                  source);
             }
+
+            if (wp != NULL)
+              {
+                g_debug ("%s: sink for %s window (id %u) has gone, removing",
+                        G_STRFUNC, wp->stream == NULL ? "preview" : "output",
+                        wp->window_id);
+
+                wp->removing = TRUE;
+                /*
+                xid = wp->window_id;
+                stream = wp->stream;
+                */
+                _remove_defunct_preview_sinks (engine, TRUE);
+                _remove_defunct_output_sinks (engine);
+
+                /* let's try recreating a new xvimagesink */
+                /*
+                if (is_preview)
+                  {
+                    tp_stream_engine_add_preview_window (engine, xid, &error);
+                  }
+                else
+                  {
+                    g_debug ("adding new output window with xid %d and stream %p", xid, stream);
+                    tp_stream_engine_stream_set_output_window (stream, xid,
+                        &error);
+                  }
+                */
+                /* let's shutdown the video stream */
+                close_all_video_streams (engine, error->message);
+              }
+          }
+        else
+          {
+            GSList *i;
+
+            g_debug ("%s: got an error on the video pipeline: %s", G_STRFUNC,
+                error->message);
+            g_debug ("%s: will teardown video pipeline and try a new one",
+                G_STRFUNC);
+
+            close_all_video_streams (engine, error->message);
+
+            g_debug ("%s: destroying video pipeline", G_STRFUNC);
+
+            for (i = priv->output_windows; i; i = i->next)
+              ((WindowPair *) i->data)->removing = TRUE;
+
+            for (i = priv->preview_windows; i; i = i->next)
+              ((WindowPair *) i->data)->removing = TRUE;
+
+            _remove_defunct_preview_sinks (engine, FALSE);
+            _remove_defunct_output_sinks (engine);
+            gst_element_set_state (priv->pipeline, GST_STATE_NULL);
+            gst_object_unref (priv->pipeline);
+            priv->pipeline = NULL;
+
+            g_debug ("%s: Creating new pipeline", G_STRFUNC);
+            priv->pipeline = tp_stream_engine_get_pipeline (engine);
+          }
+        g_error_free (error);
+        break;
+      case GST_MESSAGE_WARNING:
+        gst_message_parse_warning (message, &error, NULL);
+        g_warning ("%s: got warning: %s", G_STRFUNC, error->message);
+        g_error_free (error);
+        break;
+      case GST_MESSAGE_STATE_CHANGED:
+        {
+          GstState new_state;
+          GstState old_state;
+          GstState pending_state;
+          gst_message_parse_state_changed (message, &old_state, &new_state,
+              &pending_state);
+          /*
+          g_debug ("State changed for %s refcount %d old:%s new:%s pending:%s",
+              name,
+              GST_OBJECT_REFCOUNT_VALUE (source),
+              gst_element_state_get_name (old_state),
+              gst_element_state_get_name (new_state),
+              gst_element_state_get_name (pending_state));
+          */
+          if (source == priv->pipeline)
+            {
+              if (new_state == GST_STATE_PLAYING)
+                {
+                  g_debug ("Pipeline is playing, setting to TRUE");
+                  priv->pipeline_playing = TRUE;
+                  _add_pending_preview_windows (engine);
+                }
+              else
+                {
+                  g_debug ("Pipeline is not playing, setting to FALSE");
+                  priv->pipeline_playing = FALSE;
+                }
+            }
+          break;
         }
+      default:
+        break;
     }
 
+  g_free (name);
   return TRUE;
 }
 
@@ -611,33 +1012,157 @@ bus_sync_handler (GstBus *bus, GstMessage *message, gpointer data)
 {
   TpStreamEngine *engine = TP_STREAM_ENGINE (data);
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (engine);
-  WindowPair *wp;
+  GstElement *element;
+  WindowPair *wp = NULL;
 
- if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ELEMENT)
+  if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ELEMENT)
     return GST_BUS_PASS;
 
   if (!gst_structure_has_name (message->structure, "prepare-xwindow-id"))
     return GST_BUS_PASS;
 
-  g_debug ("got prepare-xwindow-id message");
+  element = GST_ELEMENT (GST_MESSAGE_SRC (message));
 
-  wp = _window_pairs_find_by_sink (priv->output_windows,
-      GST_ELEMENT (GST_MESSAGE_SRC (message)));
+  g_debug ("got prepare-xwindow-id message from %s",
+      gst_element_get_name (element));
 
-  if (wp == NULL)
-    wp = _window_pairs_find_by_sink (priv->preview_windows,
-        GST_ELEMENT (GST_MESSAGE_SRC (message)));
+  while (element != NULL)
+    {
+      wp = _window_pairs_find_by_sink (priv->output_windows, element);
+
+      if (wp == NULL)
+        wp = _window_pairs_find_by_sink (priv->preview_windows, element);
+
+      if (wp != NULL)
+        break;
+
+      element = GST_ELEMENT_PARENT (element);
+    }
 
   if (wp == NULL)
     return GST_BUS_PASS;
 
-  gst_x_overlay_set_xwindow_id (
-      GST_X_OVERLAY (GST_MESSAGE_SRC (message)), wp->window_id);
+  g_debug ("Giving video sink %p window id %d", wp->sink, wp->window_id);
+  gst_x_overlay_set_xwindow_id (GST_X_OVERLAY (GST_MESSAGE_SRC (message)),
+      wp->window_id);
+
+  if (g_object_has_property (G_OBJECT (GST_MESSAGE_SRC (message)), "sync"))
+    {
+      g_debug ("setting sync to FALSE");
+      g_object_set (G_OBJECT (GST_MESSAGE_SRC (message)), "sync", FALSE, NULL);
+    }
 
   gst_message_unref (message);
+
   return GST_BUS_DROP;
 }
 
+static void
+_create_pipeline (TpStreamEngine *obj)
+{
+  TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
+  GstElement *videosrc = NULL;
+  GstElement *tee;
+#ifndef MAEMO_OSSO_SUPPORT
+  GstElement *tmp;
+#endif
+  GstBus *bus;
+  GstCaps *filter = NULL;
+  GstElement *fakesink;
+  const gchar *elem;
+  const gchar *caps_str;
+
+  priv->pipeline = gst_pipeline_new (NULL);
+  tee = gst_element_factory_make ("tee", "tee");
+  fakesink = gst_element_factory_make ("fakesink", NULL);
+
+  if ((elem = getenv ("FS_VIDEO_SRC")) || (elem = getenv ("FS_VIDEOSRC")))
+    {
+      g_debug ("making video src with pipeline \"%s\"", elem);
+      videosrc = gst_parse_bin_from_description (elem, TRUE, NULL);
+      g_assert (videosrc);
+      gst_element_set_name (videosrc, "videosrc");
+    }
+  else
+    {
+#ifdef MAEMO_OSSO_SUPPORT
+      videosrc = gst_element_factory_make ("gconfv4l2src", NULL);
+#else
+      videosrc = gst_element_factory_make ("gconfvideosrc", NULL);
+
+      if (videosrc == NULL)
+        videosrc = gst_element_factory_make ("v4l2src", NULL);
+
+      if (videosrc == NULL)
+        videosrc = gst_element_factory_make ("v4lsrc", NULL);
+#endif
+
+      if (videosrc != NULL)
+        g_debug ("using %s as video source", GST_ELEMENT_NAME (videosrc));
+      else
+        g_debug ("failed to create video source");
+    }
+
+  if ((caps_str = getenv ("FS_VIDEO_SRC_CAPS")) || (caps_str = getenv ("FS_VIDEOSRC_CAPS")))
+    {
+      filter = gst_caps_from_string (caps_str);
+    }
+
+  if (!filter)
+    {
+      filter = gst_caps_new_simple ("video/x-raw-yuv",
+           "width", G_TYPE_INT, 176,
+           "height", G_TYPE_INT, 144,
+           "framerate", GST_TYPE_FRACTION, 15, 1,
+#ifndef MAEMO_OSSO_SUPPORT
+           "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('I', '4', '2', '0'),
+#endif
+           NULL);
+    }
+  else
+    {
+      g_debug ("applying custom caps '%s' on the video source\n", caps_str);
+    }
+
+  gst_bin_add_many (GST_BIN (priv->pipeline), videosrc, tee, fakesink,
+      NULL);
+
+#ifndef MAEMO_OSSO_SUPPORT
+  tmp = gst_element_factory_make ("videorate", NULL);
+  if (tmp != NULL)
+    {
+      g_debug ("linking videorate");
+      gst_bin_add (GST_BIN (priv->pipeline), tmp);
+      gst_element_link (videosrc, tmp);
+      videosrc = tmp;
+    }
+
+  tmp = gst_element_factory_make ("ffmpegcolorspace", NULL);
+  if (tmp != NULL);
+    {
+      g_debug ("linking ffmpegcolorspace");
+      gst_bin_add (GST_BIN (priv->pipeline), tmp);
+      gst_element_link (videosrc, tmp);
+      videosrc = tmp;
+    }
+#endif
+
+  gst_element_link (videosrc, tee);
+  gst_element_link_filtered (tee, fakesink, filter);
+  gst_caps_unref (filter);
+
+  g_debug ("Setting pipeline to playing");
+  gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
+
+  /* connect a callback to the stream bus so that we can set X window IDs
+   * at the right time, and detect when sinks have gone away */
+  bus = gst_element_get_bus (priv->pipeline);
+  gst_bus_set_sync_handler (bus, bus_sync_handler, obj);
+  priv->bus_async_source_id =
+    gst_bus_add_watch (bus, bus_async_handler, obj);
+  gst_object_unref (bus);
+}
+
 /*
  * tp_stream_engine_get_pipeline
  *
@@ -649,52 +1174,10 @@ GstElement *
 tp_stream_engine_get_pipeline (TpStreamEngine *obj)
 {
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
-  GstElement *videosrc;
-  GstElement *tee;
-  GstBus *bus;
-  GstCaps *filter;
-  GstElement *fakesink;
 
   if (NULL == priv->pipeline)
     {
-      const gchar *elem;
-
-      priv->pipeline = gst_pipeline_new (NULL);
-      tee = gst_element_factory_make ("tee", "tee");
-      fakesink = gst_element_factory_make ("fakesink", NULL);
-
-      if ((elem = getenv ("FS_VIDEO_SRC")) || (elem = getenv ("FS_VIDEOSRC")))
-        {
-          g_debug ("making video src with pipeline \"%s\"", elem);
-          videosrc = gst_parse_bin_from_description (elem, TRUE, NULL);
-          g_assert (videosrc);
-        }
-      else
-        {
-          videosrc = gst_element_factory_make ("v4l2src", NULL);
-        }
-      filter = gst_caps_new_simple(
-        "video/x-raw-yuv",
-        "width", G_TYPE_INT, 352,
-        "height", G_TYPE_INT, 288,
-        "framerate", GST_TYPE_FRACTION, 15, 1,
-        NULL);
-
-      gst_bin_add_many (GST_BIN (priv->pipeline), videosrc, tee, fakesink,
-          NULL);
-      gst_element_link (videosrc, tee);
-      gst_element_link_filtered (tee, fakesink, filter);
-      gst_caps_unref (filter);
-
-      gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
-
-      /* connect a callback to the stream bus so that we can set X window IDs
-       * at the right time, and detect when sinks have gone away */
-      bus = gst_element_get_bus (priv->pipeline);
-      gst_bus_set_sync_handler (bus, bus_sync_handler, obj);
-      priv->bus_async_source_id =
-        gst_bus_add_watch (bus, bus_async_handler, obj);
-      gst_object_unref (bus);
+      _create_pipeline (obj);
     }
 
   return priv->pipeline;
@@ -704,12 +1187,12 @@ tp_stream_engine_get_pipeline (TpStreamEngine *obj)
 static gboolean
 _remove_defunct_sinks_idle_cb (TpStreamEngine *engine)
 {
-  _remove_defunct_sinks (engine);
+  _remove_defunct_preview_sinks (engine, TRUE);
+  _remove_defunct_output_sinks (engine);
 
   return FALSE;
 }
 
-
 /**
  * tp_stream_engine_add_preview_window
  *
@@ -722,15 +1205,22 @@ _remove_defunct_sinks_idle_cb (TpStreamEngine *engine)
  *
  * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean tp_stream_engine_add_preview_window (TpStreamEngine *obj, guint window_id, GError **error)
+gboolean tp_stream_engine_add_preview_window (TpStreamEngine *obj,
+    guint window_id, GError **error)
 {
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
-  GstElement *tee, *sink, *filter, *pipeline;
   WindowPair *wp;
-  const gchar *videosink_name;
+
+  g_debug ("%s: called", G_STRFUNC);
+
+  if (priv->pipeline == NULL)
+    {
+      _create_pipeline (obj);
+    }
 
   /* try and remove any sinks which have removing = TRUE to free up Xv ports */
-  _remove_defunct_sinks (obj);
+  _remove_defunct_preview_sinks (obj, TRUE);
+  _remove_defunct_output_sinks (obj);
 
   wp = _window_pairs_find_by_window_id (priv->preview_windows, window_id);
 
@@ -741,43 +1231,20 @@ gboolean tp_stream_engine_add_preview_window (TpStreamEngine *obj, guint window_
       return FALSE;
     }
 
-  pipeline = tp_stream_engine_get_pipeline (obj);
-
-  gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
-
-  g_debug ("adding preview in window %u", window_id);
-
-  if ((videosink_name = getenv ("FS_VIDEO_SINK")) || (videosink_name = getenv("FS_VIDEOSINK")))
+  if (!priv->pipeline_playing)
     {
-      g_debug ("making video sink with pipeline \"%s\"", videosink_name);
-      sink = gst_parse_bin_from_description (videosink_name, TRUE, NULL);
+      g_debug ("%s: pipeline not playing, adding later", G_STRFUNC);
+      _window_pairs_add (&(priv->preview_windows), NULL, NULL, window_id);
+      return TRUE;
     }
   else
     {
-      g_debug ("using xvimagesink");
-      sink = gst_element_factory_make ("xvimagesink", NULL);
-      g_object_set (G_OBJECT (sink), "sync", FALSE, NULL);
+      g_debug ("%s: pipeline playing, adding now", G_STRFUNC);
+      _window_pairs_add (&(priv->preview_windows), NULL, NULL, window_id);
+      return _add_preview_window (obj, window_id, error);
     }
-
-  gst_bin_add (GST_BIN (priv->pipeline), sink);
-
-  filter = gst_element_factory_make ("ffmpegcolorspace", NULL);
-  gst_bin_add (GST_BIN (priv->pipeline), filter);
-
-  tee = gst_bin_get_by_name (GST_BIN (priv->pipeline), "tee");
-  gst_element_link_many (tee, filter, sink, NULL);
-  gst_object_unref (tee);
-
-  gst_element_set_state (pipeline, GST_STATE_PLAYING);
-
-  _window_pairs_add (&(priv->preview_windows), NULL, sink, window_id);
-
-  g_signal_emit (obj, signals[HANDLING_CHANNEL], 0);
-
-  return TRUE;
 }
 
-
 static gboolean
 bad_window_cb (TpStreamEngineXErrorHandler *handler,
                guint window_id,
@@ -816,6 +1283,7 @@ bad_window_cb (TpStreamEngineXErrorHandler *handler,
   wp->removing = TRUE;
 
   g_idle_add ((GSourceFunc) _remove_defunct_sinks_idle_cb, engine);
+  g_main_context_wakeup (NULL);
 
   return TRUE;
 }
@@ -953,7 +1421,7 @@ gboolean tp_stream_engine_remove_preview_window (TpStreamEngine *obj, guint wind
 
   wp->removing = TRUE;
 
-  _remove_defunct_sinks (obj);
+  _remove_defunct_preview_sinks (obj, TRUE);
 
   return TRUE;
 }
@@ -967,7 +1435,8 @@ tp_stream_engine_add_output_window (TpStreamEngine *obj,
 {
   TpStreamEnginePrivate *priv = TP_STREAM_ENGINE_GET_PRIVATE (obj);
 
-  _remove_defunct_sinks (obj);
+  _remove_defunct_preview_sinks (obj, TRUE);
+  _remove_defunct_output_sinks (obj);
 
   _window_pairs_add (&(priv->output_windows), stream, sink, window_id);
 
@@ -1255,3 +1724,23 @@ tp_stream_engine_emit_receiving (TpStreamEngine *obj, gchar *channel_path, guint
   g_signal_emit (G_OBJECT (obj), signals[RECEIVING], 0, channel_path,
       stream_id, TRUE);
 }
+
+/**
+ * tp_stream_engine_shutdown
+ *
+ * Implements DBus method Shutdown
+ * on interface org.freedesktop.Telepathy.StreamEngine
+ *
+ * @error: Used to return a pointer to a GError detailing any error
+ *         that occured, DBus will throw the error only if this
+ *         function returns false.
+ *
+ * Returns: TRUE if successful, FALSE if an error was thrown.
+ */
+gboolean tp_stream_engine_shutdown (TpStreamEngine *obj, GError **error)
+{
+  g_debug ("%s: Emitting shutdown signal", G_STRFUNC);
+  g_signal_emit (obj, signals[SHUTDOWN_REQUESTED], 0);
+  return TRUE;
+}
+
